@@ -1,6 +1,7 @@
 # rcov Copyright (c) 2004-2006 Mauricio Fernandez <mfp@acm.org>
 # See LEGAL and LICENSE for additional licensing information.
 
+require 'pathname'
 module Rcov
 
 class Formatter # :nodoc:
@@ -30,6 +31,10 @@ class Formatter # :nodoc:
         @callsite_analyzer = options[:callsite_analyzer]
         @comments_run_by_default = options[:comments_run_by_default]
         @callsite_index = nil
+
+        @mangle_filename = Hash.new{|h,base|
+            h[base] = Pathname.new(base).cleanpath.to_s.gsub(%r{^\w:[/\\]}, "").gsub(/\./, "_").gsub(/[\\\/]/, "-") + ".html"
+        }
     end
 
     def add_file(filename, lines, coverage, counts)
@@ -53,7 +58,7 @@ class Formatter # :nodoc:
     end
 
     def mangle_filename(base)
-        base.gsub(%r{^\w:[/\\]}, "").gsub(/\./, "_").gsub(/[\\\/]/, "-") + ".html"
+        @mangle_filename[base]
     end
 
     def each_file_pair_sorted(&b)
@@ -142,6 +147,29 @@ class Formatter # :nodoc:
         end
         index
     end
+
+    class XRefHelper < Struct.new(:file, :line, :klass, :mid, :count) # :nodoc:
+    end
+
+    def _get_defsites(ref_blocks, filename, lineno, linetext, label, &format_call_ref)
+        if @do_cross_references and 
+           (rev_xref = reverse_cross_references_for(filename, lineno))
+            refs = rev_xref.map do |classname, methodname, defsite, count|
+                XRefHelper.new(defsite.file, defsite.line, classname, methodname, count)
+            end.sort_by{|r| r.count}.reverse
+            ref_blocks << [refs, label, format_call_ref]
+        end
+    end
+
+    def _get_callsites(ref_blocks, filename, lineno, linetext, label, &format_called_ref)
+        if @do_callsites and
+           (refs = cross_references_for(filename, lineno))
+            refs = refs.sort_by{|k,count| count}.map do |ref, count|
+                XRefHelper.new(ref.file, ref.line, ref.calling_class, ref.calling_method, count)
+            end.reverse
+            ref_blocks << [refs, label, format_called_ref]
+        end
+    end
 end
 
 class TextSummary < Formatter # :nodoc:
@@ -202,6 +230,8 @@ class FullTextReport < Formatter # :nodoc:
                 case @textmode
                 when :counts
                     puts "%-70s| %6d" % [line.chomp[0,70], fileinfo.counts[i]]
+                when :gcc
+                    puts "%s:%d:%s" % [filename, i+1, line.chomp] unless fileinfo.coverage[i]
                 when :coverage
                     if @color
                         prefix = fileinfo.coverage[i] ? "\e[32;40m" : "\e[31;40m"
@@ -236,7 +266,8 @@ class TextCoverageDiff < Formatter # :nodoc:
         @mode = options[:coverage_diff_mode]
         @state_file = options[:coverage_diff_file]
         @diff_cmd = options[:diff_cmd]
-        super(options)
+        @gcc_output = options[:gcc_output]
+       super(options)
     end
 
     def execute
@@ -264,7 +295,7 @@ class TextCoverageDiff < Formatter # :nodoc:
         $stderr.puts <<-EOF
 Couldn't save coverage data to #{@state_file}.
 EOF
-    end
+    end                         # '
 
     require 'tempfile'
     def compare_state
@@ -275,7 +306,7 @@ EOF
             $stderr.puts <<-EOF
 Couldn't load coverage data from #{@state_file}.
 EOF
-            return
+            return              # '
         end
         if !(Array === format) or
             FORMAT_VERSION[0] != format[0] || FORMAT_VERSION[1] < format[1]
@@ -284,7 +315,7 @@ Couldn't load coverage data from #{@state_file}.
 The file is saved in the format  #{format.inspect[0..20]}.
 This rcov executable understands #{FORMAT_VERSION.inspect}.
 EOF
-            return
+            return              # '
         end
         each_file_pair_sorted do |filename, fileinfo|
             old_data = Tempfile.new("#{mangle_filename(filename)}-old")
@@ -305,7 +336,7 @@ EOF
             end
             new_data.close
 
-            diff = `#{@diff_cmd} -u #{old_data.path} #{new_data.path}`
+            diff = `#{@diff_cmd} -u "#{old_data.path}" "#{new_data.path}"`
             new_uncovered_hunks = process_unified_diff(filename, diff)
             old_data.close!
             new_data.close!
@@ -322,13 +353,21 @@ EOF
 
 EOF
         hunks.each do |offset, lines|
-            puts "### #{filename}:#{offset}"
-            if @color
+            if @gcc_output
+                lines.each_with_index do |line,i|
+                    lineno = offset + i
+                    flag = (/^!! / !~ line) ? "-" : ":"
+                    prefix = "#{filename}#{flag}#{lineno}#{flag}"
+                    puts "#{prefix}#{line[3..-1]}"
+                end
+            elsif @color
+                puts "### #{filename}:#{offset}"
                 lines.each do |line| 
                     prefix = (/^!! / !~ line) ? "\e[32;40m" : "\e[31;40m"
                     puts "#{prefix}#{line[3..-1].chomp}\e[37;40m"
                 end
             else
+                puts "### #{filename}:#{offset}"
                 puts lines
             end
         end
@@ -561,7 +600,9 @@ table.report td.text {
     border: #d0d0d0 1px solid;
 }
 
-table.report td.value {
+table.report td.value,
+table.report td.lines_total,
+table.report td.lines_code {
     text-align: right;
     border: #d0d0d0 1px solid;
 }
@@ -574,7 +615,9 @@ table.report tr.dark {
 EOS
 
     DEFAULT_OPTS = {:color => false, :fsr => 30, :destdir => "coverage",
-                    :callsites => false, :cross_references => false}
+                    :callsites => false, :cross_references => false,
+                    :validator_links => true 
+                   }
     def initialize(opts = {})
         options = DEFAULT_OPTS.clone.update(opts)
         super(options)
@@ -584,6 +627,7 @@ EOS
         @do_callsites = options[:callsites]
         @do_cross_references = options[:cross_references]
         @span_class_index = 0
+        @show_validator_links = options[:validator_links]
     end
 
     def execute
@@ -647,16 +691,18 @@ EOS
                                     a_(:href => mangle_filename(f.name)){ t_ { f.name } } 
                                 end
                             }
-                            [f.num_lines, f.num_code_lines].each do |value| 
-                                td_(:class => "value") { tt_{ value } }
+                            [[f.num_lines, "lines_total"], 
+                             [f.num_code_lines, "lines_code"]].each do |value, css_class| 
+                                td_(:class => css_class) { tt_{ value } }
                             end
-                            [f.total_coverage, f.code_coverage].each do |value|
+                            [[f.total_coverage, "coverage_total"],
+                             [f.code_coverage, "coverage_code"]].each do |value, css_class|
                                 value *= 100
                                 td_ { 
                                     table_(:cellpadding => 0, :cellspacing => 0, :align => "right") { 
                                         tr_ { 
                                             td_ {
-                                                 tt_ { "%3.1f%%" % value } 
+                                                 tt_(:class => css_class) { "%3.1f%%" % value } 
                                                  x_ "&nbsp;"
                                             }
                                             ivalue = value.round
@@ -712,17 +758,20 @@ EOS
                 x_{ format_overview(*files) }
                 hr_
                 x_{ blurb }
-                p_ {
-                    a_(:href => "http://validator.w3.org/check/referer") {
-                    img_(:src => "http://www.w3.org/Icons/valid-xhtml11",
-                         :alt => "Valid XHTML 1.1!", :height => 31, :width => 88)
+
+                if @show_validator_links
+                    p_ {
+                        a_(:href => "http://validator.w3.org/check/referer") {
+                            img_(:src => "http://www.w3.org/Icons/valid-xhtml11",
+                                 :alt => "Valid XHTML 1.1!", :height => 31, :width => 88)
+                        }
+                        a_(:href => "http://jigsaw.w3.org/css-validator/check/referer") {
+                            img_(:style => "border:0;width:88px;height:31px",
+                                 :src => "http://jigsaw.w3.org/css-validator/images/vcss",
+                                 :alt => "Valid CSS!")
+                        }
                     }
-                    a_(:href => "http://jigsaw.w3.org/css-validator/check/referer") {
-                        img_(:style => "border:0;width:88px;height:31px",
-                             :src => "http://jigsaw.w3.org/css-validator/images/vcss",
-                             :alt => "Valid CSS!")
-                    }
-                }
+                end
             }
         } }
         lines = output.pretty.to_a
@@ -760,41 +809,23 @@ EOS
         "<pre>#{result}</pre>"
     end
 
-    class XRefHelper < Struct.new(:file, :line, :klass, :mid, :count) # :nodoc:
-    end
-
     def create_cross_refs(filename, lineno, linetext)
         return linetext unless @callsite_analyzer && @do_callsites
-        ret = ""
         ref_blocks = []
-        if @do_cross_references and 
-           (rev_xref = reverse_cross_references_for(filename, lineno))
-            refs = rev_xref.map do |classname, methodname, defsite, count|
-                XRefHelper.new(defsite.file, defsite.line, classname, methodname, count)
-            end.sort_by{|r| r.count}.reverse
-            format_call_ref = lambda do |ref|
-                if ref.file
-                    where = "at #{normalize_filename(ref.file)}:#{ref.line}"
-                else
-                    where = "(C extension/core)"
-                end
-                CGI.escapeHTML("%7d   %s" % 
+        _get_defsites(ref_blocks, filename, lineno, "Calls", linetext) do |ref|
+            if ref.file
+                where = "at #{normalize_filename(ref.file)}:#{ref.line}"
+            else
+                where = "(C extension/core)"
+            end
+            CGI.escapeHTML("%7d   %s" % 
                                [ref.count, "#{ref.klass}##{ref.mid} " + where])
-            end
-            ref_blocks << [refs, "Calls", format_call_ref]
         end
-        if @do_callsites and
-           (refs = cross_references_for(filename, lineno))
-            refs = refs.sort_by{|k,count| count}.map do |ref, count|
-                XRefHelper.new(ref.file, ref.line, ref.calling_class, ref.calling_method, count)
-            end.reverse
-            format_called_ref = lambda do |ref|
-                r = "%7d   %s" % [ref.count, 
-                    "#{normalize_filename(ref.file||'C code')}:#{ref.line} " +
+        _get_callsites(ref_blocks, filename, lineno, "Called by", linetext) do |ref|
+            r = "%7d   %s" % [ref.count, 
+                "#{normalize_filename(ref.file||'C code')}:#{ref.line} " +
                     "in '#{ref.klass}##{ref.mid}'"]
-                CGI.escapeHTML(r)
-            end
-            ref_blocks << [refs, "Called by", format_called_ref]
+            CGI.escapeHTML(r)
         end
         
         create_cross_reference_block(linetext, ref_blocks)
@@ -874,17 +905,20 @@ EOS
                 x_{ body }
                 hr_
                 x_ { blurb }
-                p_ {
-                    a_(:href => "http://validator.w3.org/check/referer") {
-                    img_(:src => "http://www.w3.org/Icons/valid-xhtml10",
-                         :alt => "Valid XHTML 1.0!", :height => 31, :width => 88)
+
+                if @show_validator_links
+                    p_ {
+                        a_(:href => "http://validator.w3.org/check/referer") {
+                            img_(:src => "http://www.w3.org/Icons/valid-xhtml10",
+                                 :alt => "Valid XHTML 1.0!", :height => 31, :width => 88)
+                        }
+                        a_(:href => "http://jigsaw.w3.org/css-validator/check/referer") {
+                            img_(:style => "border:0;width:88px;height:31px",
+                                 :src => "http://jigsaw.w3.org/css-validator/images/vcss",
+                                 :alt => "Valid CSS!")
+                        }
                     }
-                    a_(:href => "http://jigsaw.w3.org/css-validator/check/referer") {
-                        img_(:style => "border:0;width:88px;height:31px",
-                             :src => "http://jigsaw.w3.org/css-validator/images/vcss",
-                             :alt => "Valid CSS!")
-                    }
-                }
+                end
             }
         } }
         # .pretty needed to make sure DOCTYPE is in a separate line
@@ -1006,6 +1040,119 @@ class HTMLProfiling < HTMLCoverage # :nodoc:
     end
 end
 
+class RubyAnnotation < Formatter # :nodoc:
+    DEFAULT_OPTS = { :destdir => "coverage" }
+    def initialize(opts = {})
+        options = DEFAULT_OPTS.clone.update(opts)
+        super(options)
+        @dest = options[:destdir]
+        @do_callsites = true
+        @do_cross_references = true
+
+        @mangle_filename = Hash.new{|h,base|
+            h[base] = Pathname.new(base).cleanpath.to_s.gsub(%r{^\w:[/\\]}, "").gsub(/\./, "_").gsub(/[\\\/]/, "-") + ".rb"
+        }
+    end
+
+    def execute
+        return if @files.empty?
+        FileUtils.mkdir_p @dest
+        each_file_pair_sorted do |filename, fileinfo|
+            create_file(File.join(@dest, mangle_filename(filename)), fileinfo)
+        end
+    end
+
+    private
+
+    def format_lines(file)
+        result = ""
+        format_line = "%#{file.num_lines.to_s.size}d"
+        file.num_lines.times do |i|
+            line = file.lines[i].chomp
+            marked = file.coverage[i]
+            count = file.counts[i]
+            result << create_cross_refs(file.name, i+1, line, marked) + "\n"
+        end
+        result
+    end
+
+    def create_cross_refs(filename, lineno, linetext, marked)
+        return linetext unless @callsite_analyzer && @do_callsites
+        ref_blocks = []
+        _get_defsites(ref_blocks, filename, lineno, linetext, ">>") do |ref|
+            ref.file.sub!(%r!^./!, '')
+            if ref.file
+                where = "at #{mangle_filename(ref.file)}:#{ref.line}"
+            else
+                where = "(C extension/core)"
+            end
+            "#{ref.klass}##{ref.mid} " + where + ""
+        end
+        _get_callsites(ref_blocks, filename, lineno, linetext, "<<") do |ref| # "
+            ref.file.sub!(%r!^./!, '')
+            "#{mangle_filename(ref.file||'C code')}:#{ref.line} " +
+                "in #{ref.klass}##{ref.mid}"
+        end
+        
+        create_cross_reference_block(linetext, ref_blocks, marked)
+    end
+
+    def create_cross_reference_block(linetext, ref_blocks, marked)
+        codelen = 75
+        if ref_blocks.empty?
+            if marked
+                return "%-#{codelen}s #o" % linetext
+            else
+                return linetext
+            end
+        end
+        ret = ""
+        @cross_ref_idx ||= 0
+        @known_files ||= sorted_file_pairs.map{|fname, finfo| normalize_filename(fname)}
+        ret << "%-#{codelen}s # " % linetext
+        ref_blocks.each do |refs, toplabel, label_proc|
+            unless !toplabel || toplabel.empty?
+                ret << toplabel << " "
+            end
+            refs.each do |dst|
+                dstfile = normalize_filename(dst.file) if dst.file
+                dstline = dst.line
+                label = label_proc.call(dst)
+                if dst.file && @known_files.include?(dstfile)
+                    ret << "[[" << label << "]], "
+                else
+                    ret << label << ", "
+                end
+            end
+        end
+        
+        ret
+    end
+    
+    def create_file(destfile, fileinfo)
+        #$stderr.puts "Generating #{destfile.inspect}"
+        body = format_lines(fileinfo)
+        File.open(destfile, "w") do |f|
+            f.puts body
+            f.puts footer(fileinfo)
+        end
+    end
+
+    def footer(fileinfo)
+        s  = "# Total lines    : %d\n" % fileinfo.num_lines
+        s << "# Lines of code  : %d\n" % fileinfo.num_code_lines
+        s << "# Total coverage : %3.1f%%\n" % [ fileinfo.total_coverage*100 ]
+        s << "# Code coverage  : %3.1f%%\n\n" % [ fileinfo.code_coverage*100 ]
+        # prevents false positives on Emacs
+        s << "# Local " "Variables:\n" "# mode: " "rcov-xref\n" "# End:\n"
+    end
+end
+
+
 end # Rcov
 
 # vi: set sw=4:
+# Here is Emacs setting. DO NOT REMOVE!
+# Local Variables:
+# ruby-indent-level: 4
+# End:
